import numpy as np
import matplotlib.pyplot as plt
import sys
if sys.version_info < (3,):
    range = xrange


# TO DO: REFACTOR AND COMBINE THESE SCRIPTS TO USE A SINGLE KALMAN FILTER/SMOOTHER SCRIPT
# Main differences between these functions are whether they treat certain matrices as
# constant or not

def univariate_KFS(y,Z,H,T,Q,R,mu):
    """ Kalman filtering and smoothing for univariate time series

    Notes
    ----------

    y = mu + Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    alpha : np.array
        Smoothed states

    V : np.array
        Variance of smoothed states
    """     

    # Filtering matrices
    a = np.zeros((T.shape[0],y.shape[0]+1)) 
    a[0][0] = np.mean(y[0:5]) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed
    L = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]))

    # Smoothing matrices
    N = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    V = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    alpha = np.zeros((T.shape[0],y.shape[0]+1)) 
    r = np.zeros((T.shape[0],y.shape[0]+1)) 

    # FORWARDS (FILTERING)
    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z,a[:,t]) - mu

        F[:,:,t] = np.dot(np.dot(Z,P[:,:,t]),Z.T) + H.ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z.T)/(F[:,:,t]).ravel()[0]

        L[:,:,t] = T - np.dot(K[:,t],Z)

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    # BACKWARDS (SMOOTHING)
    for t in reversed(range(y.shape[0])):
        if t != 0:
            L[:,:,t] = T - np.dot(K[:,t],Z)
            r[:,t-1] = np.dot(Z.T,v[t])/(F[:,:,t]).ravel()[0]
            N[:,:,t-1] = np.dot(Z.T,Z)/(F[:,:,t]).ravel()[0] + np.dot(np.dot(L[:,:,t].T,N[:,:,t]),L[:,:,t])
            alpha[:,t] = a[:,t] + np.dot(P[:,:,t],r[:,t-1])
            V[:,:,t] = P[:,:,t] - np.dot(np.dot(P[:,:,t],N[:,:,t-1]),P[:,:,t])
        else:
            alpha[:,t] = a[:,t]
            V[:,:,t] = P[:,:,t] 

    return alpha, V

def univariate_kalman(y,Z,H,T,Q,R,mu):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Filtered states

    P : np.array
        Filtered variances

    K : np.array
        Kalman Gain matrices

    F : np.array
        Signal-to-noise term

    v : np.array
        Residuals
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1)) 
    a[0][0] = np.mean(y[0:5]) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]))

    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z,a[:,t]) - mu

        F[:,:,t] = np.dot(np.dot(Z,P[:,:,t]),Z.T) + H.ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z.T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P, K, F, v

def univariate_kalman_smoother(y,Z,T,a,P,K,F,v):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    a : np.array
        Filtered states

    P : np.array
        Filtered variances

    K : np.array
        Kalman gain matrices

    F : np.array
        Signal-to-noise term

    v : np.array
        Residuals

    Returns
    ----------
    alpha : np.array
        Smoothed states

    V : np.array
        Variance of smoothed states
    """         

    N = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    L = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    V = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    alpha = np.zeros((T.shape[0],y.shape[0]+1)) 
    r = np.zeros((T.shape[0],y.shape[0]+1)) 

    for t in reversed(range(y.shape[0])):
        if t != 0:
            L[:,:,t] = T - np.dot(K[:,t],Z)
            r[:,t-1] = np.dot(Z.T,v[t])/(F[:,:,t]).ravel()[0]
            N[:,:,t-1] = np.dot(Z.T,Z)/(F[:,:,t]).ravel()[0] + np.dot(np.dot(L[:,:,t].T,N[:,:,t]),L[:,:,t])
            alpha[:,t] = a[:,t] + np.dot(P[:,:,t],r[:,t-1])
            V[:,:,t] = P[:,:,t] - np.dot(np.dot(P[:,:,t],N[:,:,t-1]),P[:,:,t])
        else:
            alpha[:,t] = a[:,t]
            V[:,:,t] = P[:,:,t] 
    return alpha, V

def univariate_kalman_fcst(y,Z,H,T,Q,R,mu,h):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Forecasted states

    P : np.array
        Variance of forecasted states
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1+h))
    a[0][0] = np.mean(y[0:5]) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1+h))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]+h))
    v = np.zeros(y.shape[0]+h)
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]+h))

    for t in range(0,y.shape[0]+h):
        if t >= y.shape[0]:
            v[t] = 0
            F[:,:,t] = 10**7
            K[:,t] = np.zeros(a.shape[0])
        else:
            v[t] = y[t] - np.dot(Z,a[:,t]) - mu
            F[:,:,t] = np.dot(np.dot(Z,P[:,:,t]),Z.T) + H.ravel()[0]
            K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z.T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P




def nl_univariate_KFS(y,Z,H,T,Q,R,mu):
    """ Kalman filtering and smoothing for univariate time series

    Notes
    ----------

    y = mu + Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    alpha : np.array
        Smoothed states

    V : np.array
        Variance of smoothed states
    """     

    # Filtering matrices
    a = np.zeros((T.shape[0],y.shape[0]+1)) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed
    L = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((1,1,y.shape[0]))

    # Smoothing matrices
    N = np.zeros((a.shape[0],a.shape[0],y.shape[0]))
    V = np.zeros((a.shape[0],a.shape[0],y.shape[0]))
    alpha = np.zeros((T.shape[0],y.shape[0])) 
    r = np.zeros((T.shape[0],y.shape[0])) 

    # FORWARDS (FILTERING)
    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z,a[:,t]) - mu[t]

        F[:,:,t] = np.dot(np.dot(Z,P[:,:,t]),Z.T) + H[t].ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z.T)/(F[:,:,t]).ravel()[0]

        L[:,:,t] = T - np.dot(K[:,t],Z)

        if t != (y.shape[0]-1):
        
            a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

            P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    # BACKWARDS (SMOOTHING)
    for t in reversed(range(y.shape[0])):
        if t != 0:
            L[:,:,t] = T - np.dot(K[:,t],Z)
            r[:,t-1] = np.dot(Z.T,v[t])/(F[:,:,t]).ravel()[0]
            N[:,:,t-1] = np.dot(Z.T,Z)/(F[:,:,t]).ravel()[0] + np.dot(np.dot(L[:,:,t].T,N[:,:,t]),L[:,:,t])
            alpha[:,t] = a[:,t] + np.dot(P[:,:,t],r[:,t-1])
            V[:,:,t] = P[:,:,t] - np.dot(np.dot(P[:,:,t],N[:,:,t-1]),P[:,:,t])
        else:
            alpha[:,t] = a[:,t]
            V[:,:,t] = P[:,:,t] 

    return alpha, V

def nl_univariate_kalman(y,Z,H,T,Q,R,mu):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Filtered states

    P : np.array
        Filtered variances

    K : np.array
        Kalman Gain matrices

    F : np.array
        Signal-to-noise term

    v : np.array
        Residuals
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1)) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((1,1,y.shape[0]))

    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z,a[:,t]) - mu[t]

        F[:,:,t] = np.dot(np.dot(Z,P[:,:,t]),Z.T) + H[t].ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z.T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P, K, F, v


def dl_univariate_KFS(y,Z,H,T,Q,R,mu):
    """ Kalman filtering and smoothing for univariate time series

    Notes
    ----------

    y = mu + Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    alpha : np.array
        Smoothed states

    V : np.array
        Variance of smoothed states
    """     

    # Filtering matrices
    a = np.zeros((T.shape[0],y.shape[0]+1)) 
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed
    L = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((1,1,y.shape[0]))

    # Smoothing matrices
    N = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    V = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    alpha = np.zeros((T.shape[0],y.shape[0]+1)) 
    r = np.zeros((T.shape[0],y.shape[0]+1)) 

    # FORWARDS (FILTERING)
    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu

        F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H.ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        L[:,:,t] = T - np.dot(K[:,t],Z[t])

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    # BACKWARDS (SMOOTHING)
    for t in reversed(range(y.shape[0])):
        if t != 0:
            L[:,:,t] = T - np.dot(K[:,t],Z[t])
            r[:,t-1] = np.dot(Z[t].T,v[t])/(F[:,:,t]).ravel()[0]
            N[:,:,t-1] = np.dot(Z[t].T,Z[t])/(F[:,:,t]).ravel()[0] + np.dot(np.dot(L[:,:,t].T,N[:,:,t]),L[:,:,t])
            alpha[:,t] = a[:,t] + np.dot(P[:,:,t],r[:,t-1])
            V[:,:,t] = P[:,:,t] - np.dot(np.dot(P[:,:,t],N[:,:,t-1]),P[:,:,t])
        else:
            alpha[:,t] = a[:,t]
            V[:,:,t] = P[:,:,t] 

    return alpha, V

def dl_univariate_kalman(y,Z,H,T,Q,R,mu):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Filtered states

    P : np.array
        Filtered variances

    K : np.array
        Kalman Gain matrices

    F : np.array
        Signal-to-noise term

    v : np.array
        Residuals
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1)) 
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]))

    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu

        F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H.ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P, K, F, v

def dl_univariate_kalman_fcst(y,Z,H,T,Q,R,mu,h):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Forecasted states

    P : np.array
        Variance of forecasted states
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1+h))
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1+h))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]+h))
    v = np.zeros(y.shape[0]+h)
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]+h))

    for t in range(0,y.shape[0]+h):
        if t >= y.shape[0]:
            v[t] = 0
            F[:,:,t] = 10**7
            K[:,t] = np.zeros(a.shape[0])
        else:
            v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu
            F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H.ravel()[0]
            K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P

def nld_univariate_KFS(y,Z,H,T,Q,R,mu):
    """ Kalman filtering and smoothing for univariate time series

    Notes
    ----------

    y = mu + Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    alpha : np.array
        Smoothed states

    V : np.array
        Variance of smoothed states
    """     

    # Filtering matrices
    a = np.zeros((T.shape[0],y.shape[0]+1)) # Initialization
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed
    L = np.zeros((a.shape[0],a.shape[0],y.shape[0]+1))
    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((1,1,y.shape[0]))

    # Smoothing matrices
    N = np.zeros((a.shape[0],a.shape[0],y.shape[0]))
    V = np.zeros((a.shape[0],a.shape[0],y.shape[0]))
    alpha = np.zeros((T.shape[0],y.shape[0])) 
    r = np.zeros((T.shape[0],y.shape[0])) 

    # FORWARDS (FILTERING)
    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu[t]

        F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H[t].ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        L[:,:,t] = T - np.dot(K[:,t],Z[t])

        if t != (y.shape[0]-1):
        
            a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

            P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    # BACKWARDS (SMOOTHING)
    for t in reversed(range(y.shape[0])):
        if t != 0:
            L[:,:,t] = T - np.dot(K[:,t],Z[t])
            r[:,t-1] = np.dot(Z[t].T,v[t])/(F[:,:,t]).ravel()[0]
            N[:,:,t-1] = np.dot(Z[t].T,Z[t])/(F[:,:,t]).ravel()[0] + np.dot(np.dot(L[:,:,t].T,N[:,:,t]),L[:,:,t])
            alpha[:,t] = a[:,t] + np.dot(P[:,:,t],r[:,t-1])
            V[:,:,t] = P[:,:,t] - np.dot(np.dot(P[:,:,t],N[:,:,t-1]),P[:,:,t])
        else:
            alpha[:,t] = a[:,t]
            V[:,:,t] = P[:,:,t] 

    return alpha, V

def nld_univariate_kalman(y,Z,H,T,Q,R,mu):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Filtered states

    P : np.array
        Filtered variances

    K : np.array
        Kalman Gain matrices

    F : np.array
        Signal-to-noise term

    v : np.array
        Residuals
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1)) 
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]))
    v = np.zeros(y.shape[0])
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]))

    for t in range(0,y.shape[0]):
        v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu[t]

        F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H[t].ravel()[0]

        K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P, K, F, v

def nld_univariate_kalman_fcst(y,Z,H,T,Q,R,mu,h):
    """ Kalman filtering for univariate time series

    Notes
    ----------

    y = Za_t + e_t         where   e_t ~ N(0,H)  MEASUREMENT EQUATION
    a_t = Ta_t-1 + Rn_t    where   n_t ~ N(0,Q)  STATE EQUATION

    Parameters
    ----------
    y : np.array
        The time series data

    Z : np.array
        Design matrix for state matrix a

    H : np.array
        Covariance matrix for measurement noise

    T : np.array
        Design matrix for lagged state matrix in state equation

    Q : np.array
        Covariance matrix for state evolution noise

    R : np.array
        Scale matrix for state equation covariance matrix

    mu : float
        Constant term for measurement equation

    Returns
    ----------
    a : np.array
        Forecasted states

    P : np.array
        Variance of forecasted states
    """         

    a = np.zeros((T.shape[0],y.shape[0]+1+h))
    P = np.ones((a.shape[0],a.shape[0],y.shape[0]+1+h))*(10**7) # diffuse prior asumed

    K = np.zeros((a.shape[0],y.shape[0]+h))
    v = np.zeros(y.shape[0]+h)
    F = np.zeros((H.shape[0],H.shape[1],y.shape[0]+h))

    for t in range(0,y.shape[0]+h):
        if t >= y.shape[0]:
            v[t] = 0
            F[:,:,t] = 10**7
            K[:,t] = np.zeros(a.shape[0])
        else:
            v[t] = y[t] - np.dot(Z[t],a[:,t]) - mu[t]
            F[:,:,t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + H[t].ravel()[0]
            K[:,t] = np.dot(np.dot(T,P[:,:,t]),Z[t].T)/(F[:,:,t]).ravel()[0]

        a[:,t+1] = np.dot(T,a[:,t]) + np.dot(K[:,t],v[t]) 

        P[:,:,t+1] = np.dot(np.dot(T,P[:,:,t]),T.T) + np.dot(np.dot(R,Q),R.T) - F[:,:,t].ravel()[0]*np.dot(np.array([K[:,t]]).T,np.array([K[:,t]]))

    return a, P


import numpy as np
from scipy import stats

class skewt(object):

    @staticmethod
    def logpdf(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        result = np.zeros(x.shape[0])
        result[x-loc<0] = np.log(2.0) - np.log(gamma + 1.0/gamma) + stats.t.logpdf(x=gamma*x[(x-loc) < 0], loc=loc[(x-loc) < 0]*gamma,df=df, scale=scale)
        result[x-loc>=0] = np.log(2.0) - np.log(gamma + 1.0/gamma) + stats.t.logpdf(x=x[(x-loc) >= 0]/gamma, loc=loc[(x-loc) >= 0]/gamma,df=df, scale=scale)
        return result

    @staticmethod
    def pdf(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        result = np.zeros(x.shape[0])
        result[x<0] = 2.0/(gamma + 1.0/gamma)*stats.t.pdf(x=gamma*x[(x-loc) < 0], loc=loc[(x-loc) < 0]*gamma,df=df, scale=scale)
        result[x>=0] = 2.0/(gamma + 1.0/gamma)*stats.t.pdf(x=x[(x-loc) >= 0]/gamma, loc=loc[(x-loc) >= 0]/gamma,df=df, scale=scale)
        return result

    @staticmethod
    def cdf(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        result = np.zeros(x.shape[0])
        result[x<0] = 2.0/(np.power(gamma,2) + 1.0)*stats.t.cdf(gamma*(x[x-loc < 0]-loc[x-loc < 0])/scale, df=df)
        result[x>=0] = 1.0/(np.power(gamma,2) + 1.0) + 2.0/((1.0/np.power(gamma,2)) + 1.0)*(stats.t.cdf((x[x-loc >= 0]-loc[x-loc >= 0])/(gamma*scale), df=df)-0.5)
        return result

    @staticmethod
    def ppf(q, df, loc=0.0, scale=1.0, gamma = 1.0):
        result = np.zeros(q.shape[0])
        probzero = skewt.cdf(x=np.zeros(1),loc=np.zeros(1),df=df,gamma=gamma)
        result[q<probzero] = 1.0/gamma*stats.t.ppf(((np.power(gamma,2) + 1.0) * q[q<probzero])/2.0,df)
        result[q>=probzero] = gamma*stats.t.ppf((1.0 + 1.0/np.power(gamma,2))/2.0*(q[q >= probzero] - probzero) + 0.5, df)
        return result

    @staticmethod
    def rvs(df, gamma, n):
        if type(n) == list:
            u = np.random.uniform(size=n[0]*n[1])
            result = np.split(u,n[0])
            return np.array(result)
        else:
            u = np.random.uniform(size=n)
            return skewt.ppf(q=u,df=df,gamma=gamma)



import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import scipy.special as sp
import matplotlib.pyplot as plt
import seaborn as sns

from ..parameter import Parameter, Parameters
from .. import inference as ifr
from .. import tsm as tsm
from .. import distributions as dst
from .. import data_check as dc

from .scores import *
from .gasmodels import *

class GASLLEV(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GENERALIZED AUTOREGRESSIVE SCORE LOCAL LEVEL MODELS ****

    Parameters
    ----------
    data : pd.DataFrame or np.array
        Field to specify the univariate time series data that will be used.

    integ : int (default : 0)
        Specifies how many time to difference the time series.

    target : str (pd.DataFrame) or int (np.array)
        Specifies which column name or array index to use. By default, first
        column/array will be selected as the dependent variable.

    dist : str
        Which distribution to use

    gradient_only : Boolean (default: True)
        If true, will only use gradient rather than second-order terms
        to construct the modified score.
    """

    def __init__(self,data,family,integ=0,target=None,gradient_only=False):

        # Initialize TSM object     
        super(GASLLEV,self).__init__('GASLLEV')

        self.integ = integ
        self.gradient_only = gradient_only
        self.param_no = 1
        self.max_lag = 1
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False

        self.data, self.data_name, self.is_pandas, self.index = dc.data_check(data,target)
        self.data_original = self.data.copy()

        for order in range(0,self.integ):
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name

        self._create_model_matrices()
        self._create_parameters()

        self.family = family
        
        self.model_name2, self.link, self.scale, self.shape, self.skewness, self.mean_transform = self.family.setup()
        self.model_name = self.model_name2 + " LLM"

        for no, i in enumerate(self.family.build_parameters()):
            self.parameters.add_parameter(i[0],i[1],i[2])
            self.parameters.parameter_list[no+1].start = i[3]

    def _create_model_matrices(self):
        """ Creates model matrices/vectors

        Returns
        ----------
        None (changes model attributes)
        """

        self.model_Y = np.array(self.data[self.max_lag:self.data.shape[0]])
        self.model_scores = np.zeros(self.model_Y.shape[0])

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        self.parameters.add_parameter('SC(1)',ifr.Normal(0,0.5,transform=None),dst.q_Normal(0,3))

    def _get_scale_and_shape(self,parm):
        """ Obtains appropriate model scale and shape parameters

        Parameters
        ----------
        parm : np.array
            Transformed parameter vector

        Returns
        ----------
        None (changes model attributes)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = parm[-1]  
                model_scale = parm[-2]
            else:
                model_shape = 0
                model_scale = parm[-1]
        else:
            model_scale = 0
            model_shape = 0 

        if self.skewness is True:
            model_skewness = parm[-3]
        else:
            model_skewness = 0

        return model_scale, model_shape, model_skewness

    def _model(self,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        parm = np.array([self.parameters.parameter_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        theta = np.zeros(self.model_Y.shape[0])
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        for t in range(0,self.model_Y.shape[0]):
            if t < self.max_lag:
                theta[t] = 0.0
            else:
                theta[t] = theta[t-1] + parm[0]*self.model_scores[t-1]

            self.model_scores[t] = self.family.score_function(self.model_Y[t],self.link(theta[t]),model_scale,model_shape,model_skewness)

        return theta, self.model_Y, self.model_scores

    def _mean_prediction(self,theta,Y,scores,h,t_params):
        """ Creates a h-step ahead mean prediction

        Parameters
        ----------
        theta : np.array
            The past predicted values

        Y : np.array
            The past data

        scores : np.array
            The past scores

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) parameters

        Returns
        ----------
        Y_exp : np.array
            Vector of past values and predictions 
        """     

        Y_exp = Y.copy()
        theta_exp = theta.copy()
        scores_exp = scores.copy()

        #(TODO: vectorize the inner construction here)      
        for t in range(0,h):
            new_value = theta_exp[-1] + t_params[0]*scores_exp[-1]
            if self.model_name2 == "Exponential GAS":
                Y_exp = np.append(Y_exp,[1.0/self.link(new_value)])
            else:
                Y_exp = np.append(Y_exp,[self.link(new_value)])
            theta_exp = np.append(theta_exp,[new_value]) # For indexing consistency
            scores_exp = np.append(scores_exp,[0]) # expectation of score is zero
        return Y_exp

    def _sim_prediction(self,theta,Y,scores,h,t_params,simulations):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        theta : np.array
            The past predicted values

        Y : np.array
            The past data

        scores : np.array
            The past scores

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) parameters

        simulations : int
            How many simulations to perform

        Returns
        ----------
        Matrix of simulations
        """     

        model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)

        sim_vector = np.zeros([simulations,h])

        for n in range(0,simulations):
            Y_exp = Y.copy()
            theta_exp = theta.copy()
            scores_exp = scores.copy()

            #(TODO: vectorize the inner construction here)  
            for t in range(0,h):
                new_value = theta_exp[-1] + t_params[0]*scores_exp[-1]

                if self.model_name2 == "Exponential GAS":
                    rnd_value = self.family.draw_variable(1.0/self.link(new_value),model_scale,model_shape,model_skewness,1)[0]
                else:
                    rnd_value = self.family.draw_variable(self.link(new_value),model_scale,model_shape,model_skewness,1)[0]

                Y_exp = np.append(Y_exp,[rnd_value])
                theta_exp = np.append(theta_exp,[new_value]) # For indexing consistency
                scores_exp = np.append(scores_exp,scores[np.random.randint(scores.shape[0])]) # expectation of score is zero

            sim_vector[n] = Y_exp[-h:]

        return np.transpose(sim_vector)

    def _summarize_simulations(self,mean_values,sim_vector,date_index,h,past_values):
        """ Summarizes a simulation vector and a mean vector of predictions

        Parameters
        ----------
        mean_values : np.array
            Mean predictions for h-step ahead forecasts

        sim_vector : np.array
            N simulation predictions for h-step ahead forecasts

        date_index : pd.DateIndex or np.array
            Dates for the simulations

        h : int
            How many steps ahead are forecast

        past_values : int
            How many past observations to include in the forecast plot

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?
        """ 

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector] - mean_values[(mean_values.shape[0]-h):(mean_values.shape[0])],0,0))
        forecasted_values = mean_values[-h-1:]
        plot_values = mean_values[-h-past_values:]
        plot_index = date_index[-h-past_values:]
        return error_bars, forecasted_values, plot_values, plot_index

    def neg_loglik(self,beta):
        theta, Y, _ = self._model(beta)
        parm = np.array([self.parameters.parameter_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y,self.link(theta),model_scale,model_shape,model_skewness)

    def plot_fit(self,intervals=False,**kwargs):
        """ Plots the fit of the model

        Returns
        ----------
        None (plots data and the fit)
        """

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            date_index = self.index[1:]
            mu, Y, scores = self._model(self.parameters.get_parameter_values())

            plt.figure(figsize=figsize)
            plt.subplot(2,1,1)
            plt.title("Model fit for " + self.data_name)

            if self.model_name2 == "Exponential GAS":
                values_to_plot = 1.0/self.link(mu)
            elif self.model_name2 == "Skewt GAS":
                t_params = self.transform_parameters()
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                additional_loc = (model_skewness - (1.0/model_skewness))*model_scale*m1
                values_to_plot = mu + additional_loc
            else:
                values_to_plot = self.link(mu)

            plt.plot(date_index,Y,label='Data')
            plt.plot(date_index,values_to_plot,label='GAS Filter',c='black')
            plt.legend(loc=2)   

            plt.subplot(2,1,2)
            plt.title("Filtered values for " + self.data_name)
            plt.plot(date_index,values_to_plot,label='GAS Filter',c='black')
            plt.legend(loc=2)   

            plt.show()              
    
    def plot_predict(self,h=5,past_values=20,intervals=True,**kwargs):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            # Retrieve data, dates and (transformed) parameters
            theta, Y, scores = self._model(self.parameters.get_parameter_values())          
            date_index = self.shift_dates(h)
            t_params = self.transform_parameters()

            # Get mean prediction and simulations (for errors)
            mean_values = self._mean_prediction(theta,Y,scores,h,t_params)

            if self.model_name2 == "Skewt GAS":
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                mean_values += (model_skewness - (1.0/model_skewness))*model_scale*m1 

            sim_values = self._sim_prediction(theta,Y,scores,h,t_params,15000)
            error_bars, forecasted_values, plot_values, plot_index = self._summarize_simulations(mean_values,sim_values,date_index,h,past_values)
            plt.figure(figsize=figsize)
            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(date_index[-h-1:], forecasted_values-pre, forecasted_values+pre,
                        alpha=alpha[count])         
            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            x = GASLLEV(integ=self.integ,family=self.family,data=self.data_original[:-h+t],gradient_only=self.gradient_only)
            x.fit(printer=False)
            if t == 0:
                predictions = x.predict(1)
            else:
                predictions = pd.concat([predictions,x.predict(1)])
        
        predictions.rename(columns={0:self.data_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            theta, Y, scores = self._model(self.parameters.get_parameter_values())          
            date_index = self.shift_dates(h)
            t_params = self.transform_parameters()

            mean_values = self._mean_prediction(theta,Y,scores,h,t_params)
            if self.model_name2 == "Skewt GAS":
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                forecasted_values = mean_values[-h:] + (model_skewness - (1.0/model_skewness))*model_scale*m1 
            else:
                forecasted_values = mean_values[-h:] 
            result = pd.DataFrame(forecasted_values)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result


import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from scipy.stats import multivariate_normal
import matplotlib.pyplot as plt
import seaborn as sns

from .. import arma
from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

from .kernels import *

class GPNARX(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GAUSSIAN PROCESS NONLINEAR AUTOREGRESSIVE (GP-NARX) MODELS ****

    Parameters
    ----------
    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.

    ar : int
        Field to specify how many AR terms the model will have.

    kernel_type : str
        One of SE (SquaredExponential), OU (Ornstein-Uhlenbeck), RQ
        (RationalQuadratic), Periodic, ARD. Defines kernel choice for GP-NARX.

    integ : int (default : 0)
        Specifies how many time to difference the time series.

    target : str (pd.DataFrame) or int (np.array)
        Specifies which column name or array index to use. By default, first
        column/array will be selected as the dependent variable.
    """

    def __init__(self,data,ar,kernel_type='SE',integ=0,target=None):

        # Initialize TSM object
        super(GPNARX,self).__init__('GPNARX')

        # Parameters
        self.ar = ar
        self.integ = integ

        if kernel_type == 'ARD':
            self.param_no = 2 + self.ar
        elif kernel_type == 'RQ':
            self.param_no = 3 + 1
        else:
            self.param_no = 3

        self.max_lag = self.ar
        self.model_name = 'GPNARX(' + str(self.ar) + ')'
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"     
        self.kernel_type = kernel_type
        self.multivariate_model = False

        # Format the data
        self.data, self.data_name, self.is_pandas, self.index = dc.data_check(data,target)
        self.data_original = self.data.copy()

        # Difference data
        for order in range(self.integ):
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name
        self.index = self.index[self.integ:len(self.index)]

        # Apply normalization
        self.data_full = self.data.copy()       
        self.data = np.array(self.data_full[self.max_lag:self.data_full.shape[0]]) # adjust for lags
        self._norm_mean = np.mean(self.data)
        self._norm_std = np.std(self.data)  
        self.data = (self.data - self._norm_mean) / self._norm_std
        self.data_full = (self.data_full - self._norm_mean) / self._norm_std

        # Define parameters
        self._create_parameters()

    def _alpha(self,L):
        """ Covariance-derived term to construct expectations. See Rasmussen & Williams.

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The alpha matrix/vector
        """     

        return np.linalg.solve(np.transpose(L),np.linalg.solve(L,np.transpose(self.data)))

    def _construct_predict(self,beta,h):    
        """ Creates h-step ahead forecasts for the Gaussian process

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        h: int
            How many steps ahead to forecast

        Returns
        ----------
        - predictions
        - variance of predictions
        """             

        # Refactor this entire code
        self._start_params(beta)
        Xstart = self.X().copy()
        Xstart = [i for i in Xstart]
        predictions = np.zeros(h)
        variances = np.zeros(h)

        for step in range(0,h):
            Xstar = []

            for lag in range(0,self.max_lag):
                if lag == 0:
                    if step == 0:
                        Xstar.append([self.data[-1]])
                        Xstart[0] = np.append(Xstart[0],self.data[-1])
                    else:
                        Xstar.append([predictions[step-1]])
                        Xstart[0] = np.append(Xstart[0],predictions[step-1])
                else:
                    Xstar.append([Xstart[lag-1][-2]])
                    Xstart[lag] = np.append(Xstart[lag],Xstart[lag-1][-2])

            Kstar = self.kernel.Kstar(np.transpose(np.array(Xstar)))

            L = self._L(beta)
            
            predictions[step] = np.dot(np.transpose(Kstar),self._alpha(L))
            variances[step] = self.kernel.Kstarstar(np.transpose(np.array(Xstar))) 
            - np.dot(np.dot(np.transpose(self.kernel.Kstar(np.transpose(np.array(Xstar)))),
                np.linalg.pinv(self.kernel.K() + np.identity(self.kernel.K().shape[0])*self.parameters.parameter_list[0].prior.transform(beta[0]))),
                self.kernel.Kstar(np.transpose(np.array(Xstar)))) + self.parameters.parameter_list[0].prior.transform(beta[0])  

        return predictions, variances, predictions - 1.98*np.power(variances,0.5), predictions + 1.98*np.power(variances,0.5)

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        self.parameters.add_parameter('Noise Sigma^2',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

        if self.kernel_type == 'ARD':
            self.kernel = ARD(self.X(),np.ones(self.ar),1)

            for lag in range(self.ar):
                self.parameters.add_parameter('l lag-' + str(lag+1),ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

        else:
            self.parameters.add_parameter('l',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

            if self.kernel_type == 'SE':
                self.kernel = SquaredExponential(self.X(),1,1)
            elif self.kernel_type == 'OU':
                self.kernel = OrnsteinUhlenbeck(self.X(),1,1)
            elif self.kernel_type == 'Periodic':
                self.kernel = Periodic(self.X(),1,1)
            elif self.kernel_type == 'RQ':
                self.parameters.add_parameter('alpha',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))
                self.kernel = RationalQuadratic(self.X(),1,1,1)

        self.parameters.add_parameter('tau',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

        arma_start = arma.ARIMA(self.data,ar=self.ar,ma=0,integ=self.integ)
        x = arma_start.fit()
        arma_starting_values = arma_start.parameters.get_parameter_values()

        for parameter in range(len(self.parameters.parameter_list)):
            self.parameters.parameter_list[parameter].start = -1.0

        self.parameters.parameter_list[0].start = np.log(np.exp(np.power(arma_starting_values[-1],2)))

    def _L(self,beta):
        """ Creates cholesky decomposition of covariance matrix

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The cholesky decomposition (L) of K
        """ 

        return np.linalg.cholesky(self.kernel.K()) + np.identity(self.kernel.K().shape[0])*self.parameters.parameter_list[0].prior.transform(beta[0])

    def _start_params(self,beta):
        """ Transforms parameters for use in kernels

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        None (changes data in self.kernel)
        """

        if self.kernel_type == 'ARD':
            for lag in range(0,self.ar):
                self.kernel.l[lag] = self.parameters.parameter_list[1+lag].prior.transform(beta[1+lag])
            self.kernel.tau = self.parameters.parameter_list[-1].prior.transform(beta[-1])  
        elif self.kernel_type == 'RQ':
            self.kernel.l = self.parameters.parameter_list[1].prior.transform(beta[1])
            self.kernel.a = self.parameters.parameter_list[2].prior.transform(beta[2])          
            self.kernel.tau = self.parameters.parameter_list[3].prior.transform(beta[3])    
        else:
            self.kernel.l = self.parameters.parameter_list[1].prior.transform(beta[1])
            self.kernel.tau = self.parameters.parameter_list[2].prior.transform(beta[2])            

    def X(self):
        """ Creates design matrix of variables to use in GP regression

        Returns
        ----------
        The design matrix
        """     

        for i in range(0,self.ar):
            datapoint = self.data_full[(self.max_lag-i-1):-i-1]         
            if i==0:
                X = datapoint
            else:
                X = np.vstack((X,datapoint))
        return X

    def expected_values(self,beta):
        """ Expected values of the function given the covariance matrix and hyperparameters

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The expected values of the function
        """     

        self._start_params(beta)
        L = self._L(beta)
        return np.dot(np.transpose(self.kernel.K()),self._alpha(L))

    def variance_values(self,beta):
        """ Covariance matrix for the estimated function

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        Covariance matrix for the estimated function 
        """     

        self._start_params(beta)

        return self.kernel.K() - np.dot(np.dot(np.transpose(self.kernel.K()),np.linalg.pinv(self.kernel.K() + np.identity(self.kernel.K().shape[0])*self.parameters.parameter_list[0].prior.transform(beta[0]))),self.kernel.K()) + self.parameters.parameter_list[0].prior.transform(beta[0])

    def neg_loglik(self,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The negative log marginal logliklihood of the model
        """             
        self._start_params(beta)
        L = self._L(beta)
        return -(-0.5*(np.dot(np.transpose(self.data),self._alpha(L))) - np.trace(L) - (self.data.shape[0]/2)*np.log(2*np.pi))

    def plot_fit(self,intervals=True,**kwargs):
        """ Plots the fit of the Gaussian process model to the data

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        intervals : Boolean
            Whether to plot uncertainty intervals or not

        Returns
        ----------
        None (plots the fit of the function)
        """

        figsize = kwargs.get('figsize',(10,7))

        date_index = self.index[self.max_lag:]
        expectation = self.expected_values(self.parameters.get_parameter_values())
        posterior = multivariate_normal(expectation,self.variance_values(self.parameters.get_parameter_values()),allow_singular=True)
        simulations = 500
        sim_vector = np.zeros([simulations,len(expectation)])

        for i in range(simulations):
            sim_vector[i] = posterior.rvs()

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append([(np.percentile(i,pre)*self._norm_std + self._norm_mean) for i in sim_vector.transpose()] 
                - (expectation*self._norm_std + self._norm_mean))

        plt.figure(figsize=figsize) 

        plt.subplot(2, 2, 1)
        plt.title(self.data_name + " Raw")  
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k')

        plt.subplot(2, 2, 2)
        plt.title(self.data_name + " Raw and Expected") 
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 3)
        plt.title(self.data_name + " Raw and Expected (with intervals)")    

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 4)

        plt.title("Expected " + self.data_name + " (with intervals)")   

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.show()

    def plot_predict(self,h=5,past_values=20,intervals=True,**kwargs):

        """ Plots forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        - Error bars, forecasted_values, plot_values, plot_index
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            predictions, variance, lower, upper = self._construct_predict(self.parameters.get_parameter_values(),h) 
            full_predictions = np.append(self.data,predictions)
            full_lower = np.append(self.data,lower)
            full_upper = np.append(self.data,upper)
            date_index = self.shift_dates(h)

            # Plot values (how far to look back)
            plot_values = full_predictions[-h-past_values:]*self._norm_std + self._norm_mean
            plot_index = date_index[-h-past_values:]

            # Lower and upper intervals
            lower = np.append(full_predictions[-h-1],lower)
            upper = np.append(full_predictions[-h-1],upper)

            plt.figure(figsize=figsize)
            if intervals == True:
                plt.fill_between(date_index[-h-1:], 
                    lower*self._norm_std + self._norm_mean, 
                    upper*self._norm_std + self._norm_mean,
                    alpha=0.2)          
            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            x = GPNARX(ar=self.ar,kernel_type=self.kernel_type,integ=self.integ,data=self.data_original[:-h+t])
            if t == 0:
                x.fit(printer=False)
                predictions = x.predict(1)  
            else:
                x.fit(printer=False)
                predictions = pd.concat([predictions,x.predict(1)])

        predictions.rename(columns={0:self.data_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[-h:]
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(date_index,data*self._norm_std + self._norm_mean,label='Data')
        plt.plot(date_index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            predictions, _, _, _ = self._construct_predict(self.parameters.get_parameter_values(),h)    
            predictions = predictions*self._norm_std + self._norm_mean  
            date_index = self.shift_dates(h)
            result = pd.DataFrame(predictions)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result





from .arma import *
from .var import *
from .gas import *
from .garch import *
from .gpnarx import *
from .inference import *
from .ssm import *
from .covariances import *
from .output import *
from .tests import *
